program generatesourcemap;
// https://sokra.github.io/source-map-visualization/#custom
{$APPTYPE CONSOLE}
{$R *.res}

uses
  System.SysUtils, classes;

const
  Base64Chars =
    'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';

function EncodeBase64VLQ(i: NativeInt): String;
{ Convert signed number to base64-VLQ:
  Each base64 has 6bit, where the most significant bit is the continuation bit
  (1=there is a next base64 character).
  The first character contains the sign bit in the last bit (1=negative)
  and the 5 least significant bits of the number.
  For example:
  A = 0 = %000000 => 0
  B = 1 = %000001 => -0
  C = 2 = %000010 => 1
  iF = 34 5 = %100010 %000101 = + 0001 00101 = 100101 = 37
}
  procedure RaiseRange;
  begin
    raise ERangeError.Create('EncodeBase64VLQ');
  end;

var
  digits: NativeInt;
begin
  Result := '';
  if i < 0 then
  begin
    i := -i;
    if i > (High(NativeInt) - 1) shr 1 then
      RaiseRange;
    i := (i shl 1) + 1;
  end
  else
  begin
    if i > High(NativeInt) shr 1 then
      RaiseRange;
    i := i shl 1;
  end;
  repeat
    digits := i and $1F; // %11111;  !![Kryvich] Delphi under 11.0 compatibility
    i := i shr 5;
    if i > 0 then
      inc(digits, $20);
    // %100000); // need another char -> set continuation bit
    Result := Result + Base64Chars[digits + 1];
  until i = 0;
end;

type
  Tfilemappings = class(TObject)
  public
    filename: string;
    fileindex: integer;
    lastoffset: integer;

  end;

var
  filelist: TStringList;

  // adds filename to list if not already there
  // returns the last known offset
function get_file_offset(filename: string; var fileindex: integer)
  : Tfilemappings;
var
  fm: Tfilemappings;
begin
  if not assigned(filelist) then
    filelist := TStringList.Create;

  fileindex := filelist.IndexOf(filename);
  if fileindex < 0 then
  begin
    writeln('adding source file: ', filename );
    fileindex := filelist.Add(filename);
    fm := Tfilemappings.Create;
    filelist.Objects[fileindex] := fm;
    fm.filename := filename;
    fm.lastoffset := 0;
  end
  else
    fm := Tfilemappings(filelist.Objects[fileindex]);
  fm.fileindex := fileindex;
  Result := fm;

end;

var
  // lastposnstart: integer = 0; // last time
  dummylastline: integer = 0;

function encode_one_basic(jsposn, jscol, jsfile, ewblinenum,
  ewbcolnum: integer): string;
begin
  Result := EncodeBase64VLQ(jsposn - dummylastline) + ',';
  Result := Result + EncodeBase64VLQ(jscol) + EncodeBase64VLQ(jsfile) +
    EncodeBase64VLQ(ewblinenum) + EncodeBase64VLQ(ewbcolnum) + ';';
  inc(dummylastline);

end;

function encode_one_null(jsposn: integer): string;
begin
  Result := EncodeBase64VLQ(jsposn - dummylastline) + ';';
  // + encodebase64vlq(0)+ encodebase64vlq(0)+encodebase64vlq(0)+';';
  inc(dummylastline);
end;

// var
// indextotal: integer = 0;

var
  lastgloballinenum : integer = 0;
  lastglobalfilenum : integer = 0;

function encode_one_mapping(filename: string; jslinenum, jsposn: integer;
  linenum, column: integer): string;
var
  fileindex, lastoffset: integer;
  fm: Tfilemappings;

begin
  try

    fm := get_file_offset(filename, fileindex);

 //   lastoffset := fm.lastoffset;
    lastoffset := lastgloballinenum;

    // inc(indextotal);
    Result := encode_one_basic(jslinenum, { - lastposnstart }
      // cursor position
      jsposn, // column
      fm.fileindex - lastglobalfilenum, // file
      linenum - lastoffset - 1, // line #
      column); // col #

    writeln(fm.fileindex, ' file last offset ', lastoffset, ' = ',
      fm.lastoffset);

    writeln('Encode 1 ', jslinenum, ' jsposn ',
      // cursor position
      jsposn, // column
      ' file index ', fm.fileindex, // file
      ' offset ', linenum - lastoffset - 1, // line #
      ' column ', column); // col #

    // lastposnstart := jsposn;
    fm.lastoffset := linenum - 1;
    lastgloballinenum := linenum - 1;
    lastglobalfilenum := fm.fileindex;
  except
    on E: Exception do
      writeln(E.ClassName, ': ', E.Message);
  end;
end;

procedure writemapfile(destdir: string; javascriptfilename: string);
var
  fil: TStringList;
  jsfullfilename: string;
  mapfilename: string;
  debugfilename: string;
  mappings: string;
  posn, i, j: integer;
  s: string;
  srcfilename: string;
  linenum: integer;
  posnstart: integer;
  lineend: string;
  bigtext: ansistring;
  jslinenum: integer;
  jsline: string;
begin
  lineend := #13;

  filelist := TStringList.Create;
  filelist.LineBreak := lineend;
  mapfilename := changefileext(javascriptfilename, '.map');
  debugfilename := changefileext(javascriptfilename, '.debug');
  writeln('creating MAP file : ' + mapfilename);

  fil := TStringList.Create;

  jsfullfilename := destdir + '\' + javascriptfilename;
  fil.LoadFromFile(jsfullfilename);

  // convert to ANSI
  // bigtext := bfile.ToString;
  bigtext := fil.Text;

  if Pos('sourceMappingURL', bigtext) < 1 then
  begin
    bigtext := '//# sourceMappingURL=' + mapfilename + fil.LineBreak + bigtext;

    writeln('adding map information to javascript file' + #13);
  end;

  fil.Clear;
  fil.Text := bigtext;
  fil.SaveToFile(destdir + '\' + javascriptfilename);

  mappings := '';

  for jslinenum := 0 to fil.Count - 1 do
  begin
    jsline := fil[jslinenum];

    // find line references
    posn := Pos('mappit_1("', jsline);
    if posn > 0 then
    begin

      // found mappit_1(
      s := copy(jsline, posn, 20);

      // get start and end of string
      i := Pos('"', jsline, posn) + 1;

      j := Pos('")', jsline, i);
      // this is the text
      s := copy(jsline, i, j - i);

      i := Pos(':', s);
      srcfilename := copy(s, 1, i - 1);

      // now srfilename and i points to integer of line num

      if i > 0 then
      begin
        writeln('');

        linenum := strtointdef(copy(s, i + 1), -1);
        if linenum >= 1 then
        begin

          // get start of next
          posnstart := 0;
          posnstart := 0;
          writeln('REFERENCE: ', srcfilename, ' : ', linenum, ' : @', posnstart,
            ' : ', copy(jsline, posnstart, 10));
          mappings := mappings + encode_one_mapping(srcfilename, jslinenum,
            posnstart, linenum, 0);
        end;

      end;

    end
    else
      mappings := mappings + encode_one_null(jslinenum);

  end;

  fil.Clear;
  fil.Text := '{ ' + #13;
  fil.Text := fil.Text + '"version": 3,' + #13;
  fil.Text := fil.Text + '"file" : "' + mapfilename + '",' + #13;
  fil.Text := fil.Text + '"sourceRoot" : "source", '+#13;

  fil.Text := fil.Text + ' "sources": [ ' + #13;

  s := '';
  for i := 0 to filelist.Count - 1 do
  begin
    fil.Text := fil.Text + '"' + filelist[i] + '"';
    if i = filelist.Count - 1 then
      fil.Text := fil.Text + #13
    else
      fil.Text := fil.Text + ',' + #13;
  end;
  fil.Text := fil.Text + '],' + #13;

  fil.Text := fil.Text + '"mappings": "' + mappings + '"' + #13 + '}';

  fil.SaveToFile(destdir + '\' + mapfilename);
end;

var
  filename: string;
  path: string;
  combo : string;

begin

  // I stored the files in z:\sourcemap
  if paramcount > 0 then begin
    combo := ParamStr( 1 );

    path := ExtractFilePath( combo );

    filename := ExtractFilename( combo );
    if ExtractFileExt( filename ) <> '.js' then begin
       writeln('Fileextension must be .js' );
       exit;
    end;

    if not FileExists( combo ) then begin
      writeln('Could not find ' + combo );
      exit;
    end;
    writemapfile( path, filename ) ;
  end else
    writeln( 'generatesourcemap z:\sourcemap\test.js');

end.
